#include <assert.h>

#include "heuristics.h"
#include "utility.h"
#include "output.h"

#include "list_multiset.h"

multiset *labels;

/* put these outside the function, as eval
 * will check to make sure values were assigned.
 * (we may have unreachable goal atoms)
 */

#define SUP_BY_INITIAL_STATE -2

static int *earliest_use = NULL;

static multiset relaxed_plan;
/* temp label for atoms to be supported */
static multiset to_support, supported;

static int *effect_supporters;

float *prev_atom_cost;
float *prev_action_cost;

int lnum_level_one_goals;
int *llevel_one_goals;


/* attempt action this iteration? */
static Bool *aa = NULL;

/* The current goal in an agenda-based algo */
State lcurrent_goals;
float gcost = 0.0;

/* forward decs */

void print_helpful();
void print_action_label(multiset_pointer lp);

void reset_label_info() {

  int i;
  
  gcost = 0.0;
  lnum_level_one_goals = 0;

  set_empty(&relaxed_plan);
  set_empty(&to_support);
  set_empty(&supported);
 
  for(i = 0; i < gnum_ft_conn; i++) {
    earliest_use[i] = INFINITY;
    gft_conn[i].is_goal = FALSE;
  }
}

float eval(State *S, int *num_acts, Bool collect_helpful, Bool use_assumptions, State *current_goals) {

  static Bool first_call = TRUE;

  int i, op, fact, pc, count;
  void *ref;

  /*
    printf("-- start --");
    print_state(S);
    printf("\n-- end --\n");
    print_state(lcurrent_goals);
  */

  if(first_call) {
    init_multiset(&relaxed_plan);
    init_multiset(&to_support);
    init_multiset(&supported);
    llevel_one_goals = calloc(gnum_ft_conn, sizeof(int));
    earliest_use = calloc(gnum_ft_conn, sizeof(int));
    labels = calloc(gnum_ft_conn, MULTISET_SIZE);
    for(i = 0; i < gnum_ft_conn; i++) {
      init_multiset( labels + i );
    }
    effect_supporters = ( int * ) calloc( gnum_ft_conn, sizeof( int ) );

    aa = calloc(gnum_ef_conn, sizeof(Bool));
    prev_atom_cost = calloc(gnum_ft_conn, sizeof(float));
    prev_action_cost = calloc(gnum_ef_conn, sizeof(float));
    gA = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    gnum_A = 0;

    first_call = FALSE;
  }

  /* (re)initialize stuff */

  source_to_dest( &lcurrent_goals, current_goals );  

  gevaluated_states++;

  reset_label_info();

  if (gcmd_line.hadd_de) {
    compute_hadd(S, use_assumptions);
  }
  
  else if (gcmd_line.hsa) {
    compute_hsa(S, use_assumptions);
    /* and take union over goal atoms. */
    for(i = 0; i < lcurrent_goals.num_F; i++) {
  
      if ( (effect_supporters[lcurrent_goals.F[i]] != SUP_BY_INITIAL_STATE) && 
	   (gef_conn[effect_supporters[lcurrent_goals.F[i]]].level == INFINITY) ) {
	*num_acts = INFINITY;
	return INFINITY;
      }
      multiset_union(&relaxed_plan, &labels[ lcurrent_goals.F[i] ]);
    }
  }
    
  for(i = 0; i < lcurrent_goals.num_F; i++) {
    if((gft_conn[lcurrent_goals.F[i]].level == 1) && (!gft_conn[lcurrent_goals.F[i]].is_goal)) {
      llevel_one_goals[lnum_level_one_goals++] = lcurrent_goals.F[i];
    }    
    gft_conn[lcurrent_goals.F[i]].is_goal = TRUE;
  }


  for(i = 0; i < lcurrent_goals.num_F; i++) {
    add_elem(&to_support, lcurrent_goals.F[i]);
  }

  while(remove_one_and_return(&to_support, &fact)) {
    if(effect_supporters[fact] == INFINITY) {
      *num_acts = INFINITY;
      return INFINITY;
    }
    if(effect_supporters[fact] != SUP_BY_INITIAL_STATE) {

      if(gcmd_line.hadd_de) {
	/* if we want the h_a relaxed plan, we add the operators */
	/* in the plan, we only want operators */
	add_elem(&relaxed_plan, gef_conn[effect_supporters[fact]].op);	
	for(i = 0; i < gef_conn[effect_supporters[fact]].num_A; i++) {
	  add_elem(&supported, gef_conn[effect_supporters[fact]].A[i]);
	}
	/*
	  printf("adding to relaxed plan: ");
	  print_op_name(gef_conn[effect_supporters[fact]].op);
	  printf(" as supporter of ");
	  print_ft_name(fact);
	  printf("\n");
	*/
      }

      /*         
	   printf("operator ");
	   print_op_name(gef_conn[effect_supporters[fact]].op);
	   printf(" has %d precons:\n",gef_conn[effect_supporters[fact]].num_PC );
	   for(i = 0; i < gef_conn[effect_supporters[fact]].num_PC; i++) {
	   printf("\t");
	   print_ft_name(gef_conn[effect_supporters[fact]].PC[i]);
	   }
	   printf("\n");
      */


      /* we use information from effects in both cases to order the helpful actions */
      for(i = 0; i < gef_conn[effect_supporters[fact]].num_PC; i++) {

	/*	
	  printf("pc: ");
	  print_ft_name(gef_conn[effect_supporters[fact]].PC[i]);
	  printf(" supported by ");
	  if(effect_supporters[gef_conn[effect_supporters[fact]].PC[i]] != SUP_BY_INITIAL_STATE)
	  print_op_name(gef_conn[effect_supporters[gef_conn[effect_supporters[fact]].PC[i]]].op);
	  else {
	  printf("(start state)");
	  }
	  printf("\n");
	*/

	pc = gef_conn[effect_supporters[fact]].PC[i];
	/*	if ( !use_assumptions && gop_conn[gef_conn[pc].op].action->is_assumption )
		continue; */
	if(!get_count(&supported, pc)){
	  add_elem(&to_support, pc);
	}
	if(LESS(gef_conn[effect_supporters[fact]].level, earliest_use[pc])) {
	  earliest_use[pc] = gef_conn[effect_supporters[fact]].level;
	}
	if((gft_conn[pc].level == 1) && (!gft_conn[pc].is_goal)){
	  llevel_one_goals[lnum_level_one_goals++] = pc;
	  gft_conn[pc].is_goal = TRUE;
	}
      }

    } /* end if */
  } /* end while */

  *num_acts = 0;
  init_iterator(&relaxed_plan, &ref);
  while(ref) {
    op = get_val(&relaxed_plan, &ref, &count);
    
    if ( !use_assumptions && gop_conn[op].action->is_assumption )
      gcost += (get_cost(op) * count * 100);
    else
      gcost += (get_cost(op) * count);
    *num_acts += count;
  }

  if(collect_helpful) {    
    collect_H_info_label();
  }

  if ( gcmd_line.display_info == 123 ) {
    /* DEBUG purposes */
    printf("---- ONE EVALUATION ----\n");
    print_relaxed_plan();
    print_helpful();
  }

  return gcost;
}
 
void compute_hsa(State *S, Bool use_ass) {

  /* label to store preconds */
  static multiset full_preconds;

  static Bool first_call = TRUE;

  int i, j, k, max_level, op, count;
  float preconds_cost = 0;
  Bool fp = FALSE;

  void *ref;
  
  if(first_call) {
    init_multiset(&full_preconds);
    first_call = FALSE;
  }

  /* get rid of old applicable actions */
  for ( i = 0; i < gnum_A; i++ ) {
    gop_conn[gA[i]].is_in_A = FALSE;
  }
  gnum_A = 0;

  /* all atom costs infinite,labels empty */
  for(i = 0; i < gnum_ft_conn; i++) {
    gft_conn[i].level = INFINITY;
    prev_atom_cost[i] = INFINITY;
    set_empty(&(labels[i]));
  }
    
  for(i = 0; i < gnum_ef_conn; i++) {
    aa[i] = FALSE;

    /*  if ( use_ass || !(gop_conn[gef_conn[i].op].action->is_assumption) )*/
     if(gef_conn[i].num_PC == 0) {
      /* effect always applicable */
      aa[i] = TRUE;

      if (! gop_conn[gef_conn[i].op].is_in_A ) {
	gop_conn[gef_conn[i].op].is_in_A = TRUE;
	gA[gnum_A++] = gef_conn[i].op;
      }
    }

    gef_conn[i].level = INFINITY;
    prev_action_cost[i] = INFINITY;
  }

  /* start atom costs 0, labels already empty */
  for(i = 0; i < S->num_F; i++) {
    gft_conn[S->F[i]].level = 0;
    effect_supporters[S->F[i]] = SUP_BY_INITIAL_STATE;
    prev_atom_cost[S->F[i]] = 0;

    for(j = 0; j < gft_conn[S->F[i]].num_PC; j++) {
      aa[gft_conn[S->F[i]].PC[j]] = TRUE;
    }
  }

  do {
    fp = TRUE;
    for(i = 0; i < gnum_ef_conn; i++) {
      /*  if ( !use_ass && gop_conn[gef_conn[i].op].action->is_assumption )
	  continue; */
      if(aa[i]) {
	aa[i] = FALSE;
	set_empty(&full_preconds);
	max_level = 0;

	for( j = 0; j < gef_conn[i].num_PC; j++ ) {
	  if( gft_conn[gef_conn[i].PC[j]].level == INFINITY ) {
	    break;
	  } 
	  else {
	    /*  printf("Adding pc ");
		print_ft_name(gef_conn[i].PC[j]);
		printf(" (supporter: ");
		print_op_name(gef_conn[effect_supporters[gef_conn[i].PC[j]]].op);
		printf(", #%d) \n",effect_supporters[gef_conn[i].PC[j]] ); 
	    */


	    if(gft_conn[gef_conn[i].PC[j]].level > max_level) {
	      max_level = gft_conn[gef_conn[i].PC[j]].level;
	    }
	    multiset_union(&full_preconds, &(labels[gef_conn[i].PC[j]]));
	  }
	}

	if(j == gef_conn[i].num_PC) {
	  preconds_cost = 0;
	  init_iterator(&full_preconds, &ref);
	  /*	  printf("summing label costs.\n"); */
	  while(ref) {
	    op = get_val(&full_preconds, &ref, &count);
	    preconds_cost += (get_cost(op) * count);

	    /*
	      printf("\top: ");
	      print_op_name(op);
	      printf(" with cost %f and count %d\n", get_cost(op), count);  */
	  }
	  
	  if(LESS(preconds_cost, prev_action_cost[i])) {
	    gef_conn[i].level = max_level;
	    if( gef_conn[i].level == 0 ) {
	      /* 	      if ( !use_ass && gop_conn[gef_conn[i].op].action->is_assumption )
			      continue;*/
	      if (! gop_conn[gef_conn[i].op].is_in_A ) {
		gop_conn[gef_conn[i].op].is_in_A = TRUE;
		gA[gnum_A++] = gef_conn[i].op;
	      }
	    }
	    
	    prev_action_cost[i] = preconds_cost;
	    increase_count(&full_preconds, gef_conn[i].op);
	    preconds_cost += get_cost(gef_conn[i].op);

	    /*
	      printf("Cost of effect %d, operator ", i);
	      print_op_name(gef_conn[i].op);
	      printf(" is %f\n", get_cost(gef_conn[i].op));*/
	    
	    for(j = 0; j < gef_conn[i].num_A; j++) {
	      if(LESS(preconds_cost, prev_atom_cost[gef_conn[i].A[j]])) {
		prev_atom_cost[gef_conn[i].A[j]] = preconds_cost;

		/*
		  printf("Set atom cost of ");
		  print_ft_name(gef_conn[i].A[j]);
		  printf(" to %f\n", prev_atom_cost[gef_conn[i].A[j]]);
		*/

		copy_multiset(&full_preconds, &(labels[gef_conn[i].A[j]]));
		gft_conn[gef_conn[i].A[j]].level = gef_conn[i].level + 1;
		effect_supporters[gef_conn[i].A[j]] = i;
		for(k = 0; k < gft_conn[gef_conn[i].A[j]].num_PC; k++){
		  aa[gft_conn[gef_conn[i].A[j]].PC[k]] = TRUE;
		}
	      }
	    }
	    fp = FALSE;
	  }
	}
      }
    }
  } while (! fp);
}

void print_action_label(multiset_pointer lp) {

  void *ref;
  int count, op;
  init_iterator(lp, &ref);
  while(ref) {
    op = get_val(&relaxed_plan, &ref, &count);
    print_op_name(op);
    printf("(%d times)", count);
    printf(", ");
  }
  printf("\n\n");
}

void print_relaxed_plan() {
  print_action_label(&relaxed_plan);
}

void print_helpful() {

  int i;

  printf("num helpful = %d\n", gnum_H);

  for(i = 0; i < gnum_H; i++) {
    print_op_name(gH[i]); printf("(lev. %d, cost %4f)", 
				 gef_conn[gop_conn[gH[i]].E[0]].level,
				 get_cost(gH[i]));
    printf(", "); 
  }  

  printf("\n\n");
}


void compute_hadd(State *S, Bool use_ass) {

  /* attempt action this iteration? */

  static Bool first_call = TRUE;

  int i, j, k, max_level;
  float preconds_cost = 0;
  Bool done = FALSE;
  
  if(first_call) {

    first_call = FALSE;
  }
  
  for ( i = 0; i < gnum_A; i++ ) {
    gop_conn[gA[i]].is_in_A = FALSE;
  }

  gnum_A = 0;

  /* all atom costs infinite,labels empty */
  for(i = 0; i < gnum_ft_conn; i++) {
    prev_atom_cost[i] = INFINITY;
    gft_conn[i].level = INFINITY;
    effect_supporters[i] = INFINITY;
  }
    
  for(i = 0; i < gnum_ef_conn; i++) {
    /* if ( use_ass || !(gop_conn[gef_conn[i].op].action->is_assumption) )*/
      if(gef_conn[i].num_PC == 0) {
	aa[i] = TRUE;

	/* always applicable */
	if (! gop_conn[gef_conn[i].op].is_in_A ) {
	  gop_conn[gef_conn[i].op].is_in_A = TRUE;
	  gA[gnum_A++] = gef_conn[i].op;
	}
      }
      else {
	aa[i] = FALSE;
      }
    prev_action_cost[i] = INFINITY;
    gef_conn[i].level = INFINITY;
  }
  
  /* start atom costs 0, labels already empty */
  for(i = 0; i < S->num_F; i++) {
    prev_atom_cost[S->F[i]] = 0;
    gft_conn[S->F[i]].level = 0;
    effect_supporters[S->F[i]] = SUP_BY_INITIAL_STATE;

    for(j = 0; j < gft_conn[S->F[i]].num_PC; j++) {
      aa[gft_conn[S->F[i]].PC[j]] = TRUE;
    }
  }

  while(! done ) {

    done = TRUE;

    for(i = 0; i < gnum_ef_conn; i++) {
      if(aa[i]) {
	aa[i] = FALSE;

	preconds_cost = 0.0;
	max_level = 0;

	for( j = 0; j < gef_conn[i].num_PC; j++ ) {
	  if( gft_conn[gef_conn[i].PC[j]].level == INFINITY ) {
	    break;
	  } 
	  else {
	    preconds_cost = PLUS(prev_atom_cost[ gef_conn[i].PC[j] ], preconds_cost);
	    max_level = MAX(max_level, gft_conn[gef_conn[i].PC[j]].level);
	  }
	}
	
	if(j == gef_conn[i].num_PC) {
	  if(LESS(preconds_cost, prev_action_cost[i])) {
	    if( max_level == 0 ) {
	      /*   if ( !use_ass && gop_conn[gef_conn[i].op].action->is_assumption )
		   continue;*/
	      if (! gop_conn[gef_conn[i].op].is_in_A ) {
		gop_conn[gef_conn[i].op].is_in_A = TRUE;
		gA[gnum_A++] = gef_conn[i].op;
	      }
	    }
	  
	    prev_action_cost[i] = preconds_cost;
	    gef_conn[i].level = max_level;
	  
	    preconds_cost += get_cost(gef_conn[i].op); 
	  	  
	    for(j = 0; j < gef_conn[i].num_A; j++) {
	    
	      if(LESS(preconds_cost, prev_atom_cost[gef_conn[i].A[j]] )) {

		prev_atom_cost[gef_conn[i].A[j]] = preconds_cost;
		gft_conn[ gef_conn[i].A[j] ].level = max_level + 1;

		effect_supporters[ gef_conn[i].A[j] ] = i;
	      
		for(k = 0; k < gft_conn[gef_conn[i].A[j]].num_PC; k++){
		  aa[gft_conn[gef_conn[i].A[j]].PC[k]] = TRUE;
		}
	      }
	    }
	    done = FALSE;
	  }
	}
      }
    }
  }
}

void collect_H_info_label( void )

{

  static Bool first_call = TRUE;
  static int *H, num_H, *D, *eu_O, *sort_eu_O;
  int i, j, k, ft, ef, op, d, last_ass;

  if ( first_call ) {
    gH = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    H = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    D = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    sort_eu_O = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    eu_O = ( int * ) calloc( gnum_op_conn, sizeof( int ) );

    gnum_H = 0;
    num_H = 0;
    first_call = FALSE;
  }
  /* get rid of the previous helping actions */
  for ( i = 0; i < gnum_op_conn; i++ ) {
    gop_conn[i].is_in_H = FALSE;
  }

  for(i = 0; i < gnum_op_conn; i++) {
    eu_O[i] = INFINITY;
  }

  num_H = 0;
  /* only interested in level 1 goals for helping actions */
  /*  for ( i = 0; i < lnum_goals_at[1]; i++ ) { */
  for ( i = 0; i < lnum_level_one_goals; i++ ) {
    /*  ft = lgoals_at[1][i]; */
    ft = llevel_one_goals[i];

    if ( gcmd_line.display_info == 124 ) {
      printf("\nLevel one goal: ");
      print_ft_name(ft);
    }

    for ( j = 0; j < gft_conn[ft].num_A; j++ ) {
      ef = gft_conn[ft].A[j];

      if ( gef_conn[ef].level != 0 ) {
	continue;
      }
      op = gef_conn[ef].op;

      /* Assumptions are not helpful actions 
      if ( gop_conn[op].action->is_assumption )
	continue;
      */
      if ( gop_conn[op].is_in_H ) {
	continue;
      }

      if(LESS(earliest_use[ft], eu_O[op])) {
	eu_O[op] = earliest_use[ft];
      }
     
      if(gef_conn[ef].level == 0){
	/* actually include it as a helpful action */
	gop_conn[op].is_in_H = TRUE;
	H[num_H++] = op;
      }
    }
  }


  /* H collected; now order it
   * here: count number of goal- and subgoal facts that
   *       op deletes (with level 0 effects). order less deletes
   *       before more deletes.
   *       start from back of H, to prefer down under
   *       goals to upper goals.
   */
  gnum_H = 0;
  last_ass = 0;
  /* for each op */
  for ( i = num_H - 1; i > -1; i-- ) {
  
    d = 0;

    ef = gop_conn[H[i]].E[0];
    
    if ( gef_conn[ef].level != 0 ) continue;
    /* count the number of deletes of the op 
     * which are current goals 
     */
    for ( k = 0; k < gef_conn[ef].num_D; k++ ) {
      if ( gft_conn[gef_conn[ef].D[k]].is_goal ) d++;
    }
      
    /* next two loops do an insert sort, 
     * using #goals deleted as the criteria 
     * and now, hopefully, earliest use,
     * and also, lower cost (now commented).
     * the ASSUME actions are always the
     * last ones to be choosed */

    /* 
       NB: No assumptions can be helpful */
    if (gop_conn[H[i]].action->is_assumption) {
      j = gnum_H;
      last_ass++;
    }
    else
      if (gop_conn[H[i]].action->norm_operator->operatore->is_obs) {
	j = 0;
      }
    else
      for ( j = 0; j < gnum_H; j++ ) {
	if(D[j] == d) { 
	  if(sort_eu_O[j] > eu_O[H[i]]) {
	    break;
	  } 
	  /**	 	  else if(sort_eu_O[j] == eu_O[H[i]]) {
			  if (get_cost(gH[j]) > get_cost(H[i]))
			  break; 
			  } 
	  **/
	}

	if ( D[j] > d ) break;
      }

    /*  
	NB: in case of helpful assumptions */
    if ((j >= last_ass) && (!gop_conn[H[i]].action->is_assumption))
      j = j-last_ass;


    /** DEBUG purposes ---
	printf("\n operat ");
	print_op_name( H[i] );
	printf("will be put at place %d\n",j); */

    for ( k = gnum_H; k > j; k-- ) {
      gH[k] = gH[k-1];
      D[k] = D[k-1];
      sort_eu_O[k] = sort_eu_O[k-1];
    }
    gH[j] = H[i];
    D[j] = d;
    sort_eu_O[j] = eu_O[H[i]];
    gnum_H++;
  }

  /* display stuff */
  if ( gcmd_line.display_info == 124 ) {
    printf("\ncollected H: ");
    for ( i = 0; i < gnum_H; i++ ) 
      print_op_name( gH[i] );

  }
}



void initialize_lcurrent( void )

{
  make_state( &lcurrent_goals, gnum_ft_conn );
  lcurrent_goals.max_F = gnum_ft_conn;
}
