#include "list_multiset.h"
#include "ff.h"

#include <stdlib.h>
#include <stdio.h>

void free_nodes(node_pointer np) {
  if(np->next) {
    free_nodes(np->next);
  }
  free(np);
}

void set_empty(multiset_pointer set) {

  if(set->first) {
    free_nodes(set->first);
  }
  set->n_distinct_elems = 0;
  set->first = NULL;
}

void init_multiset(multiset_pointer set) {
  set->n_distinct_elems = 0;
  set->first = NULL;
}

void copy_multiset(multiset_pointer from, multiset_pointer to) {

  node_pointer from_np = from->first, *to_np;

  set_empty(to);

  to_np = &(to->first);

  while (from_np) {
    *to_np = malloc(MULTISET_NODE_SIZE);
    (*to_np)->val = from_np->val;
    (*to_np)->count = from_np->count;
    (*to_np)->next = NULL;
    to_np = &((*to_np)->next);
    from_np = from_np->next;
  }

  to->n_distinct_elems = from->n_distinct_elems;
}

void increase_count(multiset_pointer set, int val) {

  node_pointer prev = NULL, current = set->first, temp;

  while (current && (current->val < val)) {
    prev = current;
    current = current->next;
  }

  if(current) {
    if(current->val == val) {
      current->count++;
      return;
    }
    if(!prev) {
      temp = malloc(MULTISET_NODE_SIZE);
      temp->next = set->first;
      set->first = temp;
      temp->val = val;
      temp->count = 1;
      set->n_distinct_elems++;
      return;
    }
  }
  if(!prev) {
    set->first = malloc(MULTISET_NODE_SIZE);
    set->first->val = val;
    set->first->count = 1;
    set->first->next = NULL;
    set->n_distinct_elems++;
  }
  else {
    prev->next = malloc(MULTISET_NODE_SIZE);
    prev->next->next = current;
    prev->next->val = val;
    prev->next->count = 1;
    set->n_distinct_elems++;
  }
}

void multiset_union(multiset_pointer result, multiset_pointer operand) {

  node_pointer op_current = operand->first, temp;
  node_pointer *result_current = &(result->first);
  
  while(op_current) {

    /* here we need val from op to go before current result val */
    if(!(*result_current) || ((*result_current)->val > op_current->val)) {
      temp = malloc(MULTISET_NODE_SIZE);
      temp->val = op_current->val;
      temp->count = op_current->count;
      temp->next = *result_current;
      *result_current = temp;
      result_current = &(temp->next);
      op_current = op_current->next;
      result->n_distinct_elems++;
    }
    else if ((*result_current)->val == op_current->val) {
      if(op_current->count > (*result_current)->count) {
	(*result_current)->count = op_current->count;
      }
      op_current = op_current->next;
    }
    else {
      result_current = &((*result_current)->next);
    }
  }
}

void multiset_intersection(multiset_pointer result, multiset_pointer operand) {

  node_pointer op_current = operand->first, temp;
  node_pointer *result_current = &(result->first);

  while(*result_current) {

    if(op_current) {
      if(op_current->val <= (*result_current)->val) {
	if(op_current->val == (*result_current)->val) {
	  if( op_current->count < (*result_current)->count ) {
	    (*result_current)->count = op_current->count;
	  }
	  result_current = &((*result_current)->next);
	}
	op_current = op_current->next;
      }
      else { 
	temp = (*result_current)->next;
	free(*result_current);
	*result_current = temp;
	result->n_distinct_elems--;
      }
    }

    else {
      temp = (*result_current)->next;
      free(*result_current);
      *result_current = temp;
      result->n_distinct_elems--;
    }
  }
}


void decrease_count(multiset_pointer set, int val) {

  node_pointer prev = NULL, current = NULL;
  
  if(!set->first) return;

  current = set->first;
  
  while(current && (current->val < val)) {
    prev = current;
    current = current->next;
  }

  if(!current || (current->val != val)) {
    return;
  }

  else {
    /* if count goes down to 0 */
    if(!(--(current->count))) {
      if(prev) {
	prev->next = current->next;
      }
      else {
	set->first = current->next;
      }
      free(current);
      set->n_distinct_elems--;
    }
  }
}

int get_count(multiset_pointer set, int val) {
  
  node_pointer np = set->first;

  while(np && (np->val < val)) {
    np = np->next;
  }

  if(np && (np->val == val)) {
    return np->count;
  }

  return 0;
}

int num_distinct_elems(multiset_pointer set) {
  return set->n_distinct_elems;
}

void print_set(multiset_pointer set) {

  int i;

  printf("num elems = %d:\t", set->n_distinct_elems);

  node_pointer np = set->first;

  while(np) {
    for( i = 0; i < np->count; i++) {
      printf("%d, ", np->val);
    }
    np = np->next;
  }
  printf("\n");
}

void init_iterator(multiset_pointer set, void **track) {
  *track = set->first;
}

int get_val(multiset_pointer set, void **current, int *count) {
  int rv = ((node_pointer)*current)->val;
  *count = ((node_pointer)*current)->count;
  *current = ((node_pointer)*current)->next;
  return rv;
}


void add_elem(multiset_pointer set, int val) {
  node_pointer prev = NULL, current = set->first, temp;
  while (current && (current->val < val)) {
    prev = current;
    current = current->next;
  }
  if(current) {
    if(current->val == val) {
      /* only difference from increase_count() function */
      /* current->count++;*/
      return;
    }
    if(!prev) {
      temp = malloc(MULTISET_NODE_SIZE);
      temp->next = set->first;
      set->first = temp;
      temp->val = val;
      temp->count = 1;
      set->n_distinct_elems++;
      return;
    }
  }
  if(!prev) {
    set->first = malloc(MULTISET_NODE_SIZE);
    set->first->val = val;
    set->first->count = 1;
    set->first->next = NULL;
    set->n_distinct_elems++;
  }
  else {
    prev->next = malloc(MULTISET_NODE_SIZE);
    prev->next->next = current;
    prev->next->val = val;
    prev->next->count = 1;
    set->n_distinct_elems++;
  }
}

Bool remove_one_and_return(multiset_pointer set, int *rv) {
  node_pointer first = NULL;
  if(!set->first) {
    return FALSE;
  }
  else {
    first = set->first;
    *rv = set->first->val;
    set->first = set->first->next;
    free(first);
    set->n_distinct_elems--;
    return TRUE;
  }
}
